/*
 *  * Copyright © 2019 Pendo.io. All rights reserved.
 */

import 'dart:async';

import 'package:flutter/material.dart';
// !! we ignore using relative imports only for known issue with logger.dart static filds
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/logger.dart';
import '../pendo_sdk.dart';
import 'channel.dart';
import 'constants/constants.dart';
import 'extensions/iterable_extensions.dart';
import 'screen_scan/pendo_scan_utils.dart';
import 'tree/element_info.dart';
import 'tree/pendo_view_hierarchy_tree.dart';
import 'tree/tree_node.dart';

class PendoScan {
  // TODO: to avoid confusions with looking for a log in console we can adjust Logger method 
  // to get two parameters the TAG and the log content separately
  static const FILE_TAG = 'PendoScan';

  RetroactiveScreenData _currentRetroactiveScreenData = RetroactiveScreenData.empty();
  @visibleForTesting
  RetroactiveScreenData get currentRetroactiveScreenData => _currentRetroactiveScreenData;

  RetroactiveScreenData _previousRetroactiveScreenData = RetroactiveScreenData.empty();
  @visibleForTesting
  RetroactiveScreenData get previousRetroactiveScreenData => _previousRetroactiveScreenData;

  /// Properties
  int _depth = 1;
  late PendoViewHierarchyTree _pendoViewHierarchyTree;

  // Used for track the screen focused area
  int primaryFocusHash = 0;
  int recentScaffoldHash = 0;
  bool isPrimaryFocusNodeFound = false;
  bool isScaffoldFound = false;

  bool _scanOnlyCurrentFocusedScreen = true; // by default we want to try collect elements only for a focused area
  bool _squashNestedClickableElements = true; // by default we want to try squash nested clickable elements
  // bool _isKeyboardOpenedSkipScanNotifyied = false;

  PendoScan._internal();
  factory PendoScan() => _singleton;
  static final PendoScan _singleton = PendoScan._internal();

  void setScanningOptions(Map<String, dynamic> pendoOptions) {
    if (pendoOptions[SCAN_ONLY_CURRENT_FOCUSED_SCREEN] != null) {
      _scanOnlyCurrentFocusedScreen = pendoOptions[SCAN_ONLY_CURRENT_FOCUSED_SCREEN] as bool;
    }
    if (pendoOptions[SQUASH_NESTED_CLICKABLE_ELEMENTS] != null) {
      _squashNestedClickableElements = pendoOptions[SQUASH_NESTED_CLICKABLE_ELEMENTS] as bool;
    }
  }

  /// API
  void scanFrom(Element? element) {
    if (element != null && element.mounted) {
      // while software keyboard is open we get endless 'onFrameChange' 
      // callbacks loop that we want to skip
      // TODO: We should find a way to avoid using RootWidet as it 
      // only available from the 3.16 dart version
      /*if (_isKeyboardOpen(element.widget as RootWidget)) {
        return;
      }*/ 
      Logger.i('$FILE_TAG scan started....');
      final timeNow = DateTime.now();
      prepareForNewScan();
      visitChild(element);
      final difference = DateTime.now().difference(timeNow).inMilliseconds;
      Logger.i('$FILE_TAG scan finished in $difference milliseconds');
    } else {
      Logger.c('$FILE_TAG element not mounted. dropping internal scan');
    }
    notifySDK();
  }

  // This method returns true, if the framework FlutterView that represent the 
  // whole screen content, has bottom padding meaning there is software keyboard opened 
  /*bool _isKeyboardOpen(RootWidget rootWidget) {
    final View? view = rootWidget.child as View?;
    final FlutterView? flutterView = view?.view;
    final bool isKeyboardOpen = flutterView != null && flutterView.viewInsets.bottom > 0;
    if (isKeyboardOpen && !_isKeyboardOpenedSkipScanNotifyied) {
      Logger.i('$FILE_TAG skip scan due opened keyboard'); 
      // we want notify only once within the endless onFrameChange callbacks flow
      _isKeyboardOpenedSkipScanNotifyied = true;
    } else {
      _isKeyboardOpenedSkipScanNotifyied = false;
    }
    return isKeyboardOpen;
  }*/

  // This method clearing and preparing variables
  // for a new scan for the current screen
  @visibleForTesting
  void prepareForNewScan() {
    // reset depth
    _depth = 1;
    // reinitialize retroactive screen data for current scan
    _currentRetroactiveScreenData = ScanUtils.initializeRetroactiveScreenDataForCurrentScreen(PendoNavigationObserver.currentRouteName, PendoNavigationObserver.previousRouteName);
    // create and configure the new PendoViewHierarchyTree to collect data into
    _pendoViewHierarchyTree = PendoViewHierarchyTree(squashNestedClickableElements: _squashNestedClickableElements);
    // find primary focus that we will start collect elements info from
    isPrimaryFocusNodeFound = false;
    isScaffoldFound = false;
    recentScaffoldHash = 0;
    FocusNode? primaryFocus = WidgetsBinding.instance.focusManager.primaryFocus;
    primaryFocusHash = primaryFocus?.context?.widget.hashCode ?? 0;
  }

  // This method is for traversing the Flutter element tree and collect current screen retroactiveScreenData:
  // 1. screen special elements(childViews)that might affect screen id - BottomNavigationBar/TabBar/Drawer
  // 2. view tree that represent current screen visual hierarchy
  // 3. all current screen related texts
  @visibleForTesting
  void visitChild(Element element) {
    try {
      // for each Flutter element we should update the pendo tree depth accordingly 
      // no mater if this element itself will be part of the pendo tree
      _pendoViewHierarchyTree.updateDepth(element.depth);
      // we need the scaffold widget only to know if a Drawer opened
      if (element.widget is Scaffold) {
        isScaffoldFound = true;
      }
      // track when we get to the Flutter element tree primary focus node
      if (!isPrimaryFocusNodeFound && element.widget.runtimeType == FocusScope && primaryFocusHash == element.widget.hashCode) {
        isPrimaryFocusNodeFound = true;
        // if there is a drawer opened in the screen the primary focus will be inside it
        // if the primary focus found before scaffold means the drawer is not opened
        if (isScaffoldFound) {
          try {
            ScaffoldState scaffoldState = Scaffold.of(element);
            ScanUtils.extractDrawer(scaffoldState, _depth++, _currentRetroactiveScreenData);
          } catch (e) {
            Logger.i('$FILE_TAG scaffold was not found for current primary focus: $e');
          }
        }
      }
      
      // If the _scanOnlyCurrentFocusedScreen flag is true means -
      // we want to start collect ui/view elements info only for the current focused
      // screen content. Once we get to the primary focus node in the tree it will ensure
      // that all visited elements are the top most ui/view elements for the current screen.
      if (_scanOnlyCurrentFocusedScreen && !isPrimaryFocusNodeFound) {
        createLayoutElementInfo(element);
      } else {
        createViewElementInfo(element);
      }
      element.visitChildren(visitChild);
    } catch (e) {
      Logger.e('$FILE_TAG could\'t visit child $element, exception caught: $e');
    }
  }

  @visibleForTesting
  void notifySDK() {
    if (_currentRetroactiveScreenData.retroactiveScreenId.isEmpty) {
      Logger.c('$FILE_TAG dropping scan due to empty retroactiveScreenId');
      return;
    }
    
    if (_currentRetroactiveScreenData == _previousRetroactiveScreenData) {
      Logger.c('$FILE_TAG dropping scan due to equal arguments for ${_currentRetroactiveScreenData.retroactiveScreenId}');
      return;
    }

    _previousRetroactiveScreenData = RetroactiveScreenData.copy(_currentRetroactiveScreenData);
    // Compose the retroactive screen data and tree hierarchy under the screenScanResult 
    // and send it to the native
    var screenScanResult = {};
    screenScanResult[RETROACTIVE_SCREEN_DATA] = _currentRetroactiveScreenData.toJson();
    screenScanResult[TREE] = _pendoViewHierarchyTree.toJson();

    Logger.d('$FILE_TAG page changed with: ', object: screenScanResult);
    unawaited(PendoChannel.invokeMethod(SCREEN_CHANGED, screenScanResult));
  }

  void createLayoutElementInfo(Element element) {
    // we try to generate element info from the current Flutter element
    ElementInfo? elementInfo = ElementInfo.createLayoutElementInfo(element);
    if (elementInfo == null) {
      return;
    }

    TreeNode treeElement = TreeNode(elementInfo: elementInfo);
    _pendoViewHierarchyTree.addTreeElement(treeElement);
  }

  void createViewElementInfo(Element element) {
    // when collecting all kind of view elements we should consider 
    // 1. screen special elements(childViews)that might affect screen id
    // 2. view tree that represent current screen visual hierarchy
    // 3. all current screen related texts
    // we try to generate element info from the current Flutter element
    ElementInfo? elementInfo = ElementInfo.createViewElementInfo(element);
    if (elementInfo == null) {
      return;
    }

    if (element.widget is BottomNavigationBar) {
      ScanUtils.extractBottomNavigationBar(element.widget as BottomNavigationBar, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is TabBar) {
      ScanUtils.extractTabBar(element.widget as TabBar, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is TabBarView) {
      ScanUtils.extractTabBarView(element.widget as TabBarView, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is ListView) {
      ScanUtils.extractListView(element.widget as ListView, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is SliverFixedExtentList) {
      ScanUtils.extractSliverFixedExtentListView(element.widget as SliverFixedExtentList, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is GridPaper) {
      ScanUtils.extractGridPaperView(element.widget as GridPaper, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is Flow) {
      ScanUtils.extractFlowView(element.widget as Flow, _depth++, _currentRetroactiveScreenData);
    } else if (element.widget is DataTable) {
      ScanUtils.extractDataTableView(element.widget as DataTable, _depth++, _currentRetroactiveScreenData);
    }

    // if the generated element contain text we put it beside into retroactiveScreenData
    _currentRetroactiveScreenData.texts.addIfNotNull(elementInfo.retroElementIdentifier());
    TreeNode treeElement = TreeNode(elementInfo: elementInfo);
    _pendoViewHierarchyTree.addTreeElement(treeElement);
  }
}
