import 'package:flutter/material.dart';
// ignore: prefer_relative_imports
import '../constants/constants.dart';
import 'screen_scan_data.dart';

export 'screen_scan_data.dart' show RetroactiveScreenData;

class ScanUtils {
  
  // This method initialize retroactiveScreenData with current screen info
  // example:
  // "currentScreenName": "flutterRoot",
  // "previousScreenName": "",
  // "retroactiveScreenId": "flutterRoot", * if a page is kind of modal the appropriate __MODAL__ screen id is used with actual name as info kind
  // "childViews": []
  static RetroactiveScreenData initializeRetroactiveScreenDataForCurrentScreen(String currentRouteName, String previousRouteName) {
    RetroactiveScreenData screenData = RetroactiveScreenData(currentScreenName: currentRouteName, previousScreenName: previousRouteName, retroactiveScreenId: currentRouteName);
    
    if (isPageModal(currentRouteName)) {
      screenData.retroactiveScreenId = MODAL_SCREEN_ID;
      screenData.info = ScreenElementInfo(kind: currentRouteName);
    } 
    
    return screenData;
  }

  // This method extract data from BottomNavigationBar and place it into
  // retroactiveScreenData as childView
  // example of childElement:
  // "name": "BottomNavigationBar",
  // "depth": 20,
  // "info": {
  //   "selectedTitleText": "bottom bar navigation item label",
  //   "selectedIndex": 1,
  //   "numberOfIndexes": 3
  // }
  static void extractBottomNavigationBar(BottomNavigationBar bottomNavigationBar, int depth, RetroactiveScreenData screenData) {
    if (bottomNavigationBar.items.length > 0) {
      int index = bottomNavigationBar.currentIndex;
      int length = bottomNavigationBar.items.length;
      String? barName = bottomNavigationBar.items.elementAt(index).label;

      ScreenElementInfo screenElementInfo = ScreenElementInfo(selectedTitleText: (barName != null ? barName : ''), selectedIndex: index + 1, numberOfIndexes: length);
      ScreenChildElement screenChildElement = ScreenChildElement(name: BOTTOM_NAVIGATION_BAR, depth: depth, info: screenElementInfo);
      screenData.childViews.add(screenChildElement);
    }
  }

  // This method extract data from TabBar and place it into
  // retroactiveScreenData as childView
  // example:
  // "name": "TabBar",
  // "depth": 20,
  // "info": {
  //   "selectedTitleText": "some label or empty string",
  //   "selectedIndex": 1,
  //   "numberOfIndexes": 3
  // }
  static void extractTabBar(TabBar tabBarWidget, int depth, RetroactiveScreenData screenData) {
    var tabBarController = tabBarWidget.controller;
    if (tabBarController != null && tabBarController.length > 0) {
      int index = tabBarController.index;
      int length = tabBarController.length;
      String? tabName = (tabBarWidget.tabs.elementAt(index) as Tab).text;

      ScreenElementInfo screenElementInfo = ScreenElementInfo(selectedTitleText: (tabName != null ? tabName : ''), selectedIndex: index + 1, numberOfIndexes: length);
      ScreenChildElement screenChildElement = ScreenChildElement(name: TAB_BAR, depth: depth, info: screenElementInfo);
      screenData.childViews.add(screenChildElement);
    }
  }

  // This method extract data from Drawer and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "retroactiveScreenId": __DRAWER__,
  // "info": {
  //   "kind": "Drawer"/"EndDrawer",
  //   "depth": 3
  // }
  static void extractDrawer(ScaffoldState scaffoldState, int depth, RetroactiveScreenData screenData) {
    if (scaffoldState.isDrawerOpen) {
      screenData.retroactiveScreenId = DRAWER_SCREEN_ID;
      screenData.info = ScreenElementInfo(kind: DRAWER, depth: depth);
    }
    
    if (scaffoldState.isEndDrawerOpen) {
      screenData.retroactiveScreenId = DRAWER_SCREEN_ID;
      screenData.info = ScreenElementInfo(kind: END_DRAWER, depth: depth);
    }
  }

  // This method extract data from TabBarView and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "TabBarView",
  //   "depth": 2,
  //   "info": {
  //     "selectedIndex": 1,
  //     "numberOfIndexes": 3
  //   }
  // }]
  static void extractTabBarView(TabBarView tbView, int depth, RetroactiveScreenData screenData) {
    var tabBarController = tbView.controller;
    if (tabBarController != null && tabBarController.length > 0) {
      int index = tabBarController.index;
      int length = tabBarController.length;
      ScreenElementInfo screenElementInfo = ScreenElementInfo(selectedIndex: index + 1, numberOfIndexes: length);
      ScreenChildElement screenChildElement = ScreenChildElement(name: TABBARVIEW, depth: depth, info: screenElementInfo);
      screenData.childViews.add(screenChildElement);
    }
  }

  // This method extract data from ListView and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "ListView",
  //   "depth": 2,
  //   "info": {
  //     "childrenCount": 7
  //   }
  // }]
  static void extractListView(ListView list, int depth, RetroactiveScreenData screenData) {
    ScreenElementInfo screenElementInfo = ScreenElementInfo(childrenCount: list.semanticChildCount);
    ScreenChildElement screenChildElement = ScreenChildElement(name: LISTVIEW, depth: depth, info: screenElementInfo);
    screenData.childViews.add(screenChildElement);
  }

  // This method extract data from SliverFixedExtentList and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "SliverFixedExtentList",
  //   "depth": 2,
  //   "info": {
  //     "itemExtent": 24.0
  //   }
  // }]
  static void extractSliverFixedExtentListView(SliverFixedExtentList list, int depth, RetroactiveScreenData screenData) {
    SliverChildDelegate delegate = list.delegate;
    ScreenElementInfo screenElementInfo = ScreenElementInfo(childrenCount: delegate.estimatedChildCount, itemExtent: list.itemExtent);
    ScreenChildElement screenChildElement = ScreenChildElement(name: SILVERFIXEDEXTENTLIST, depth: depth, info: screenElementInfo);
    screenData.childViews.add(screenChildElement);
  }

  // This method extract data from GridPaper and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "GridPaper",
  //   "depth": 2,
  //   "info": {
  //    "interval": 100.0,
  //    "devision": 2,
  //    "subDevision": 5
  //   }
  // }]
  static void extractGridPaperView(GridPaper gridPaper, int depth, RetroactiveScreenData screenData) {
    ScreenElementInfo screenElementInfo = ScreenElementInfo(interval: gridPaper.interval, devision: gridPaper.divisions, subDevision: gridPaper.subdivisions);
    ScreenChildElement screenChildElement = ScreenChildElement(name: GRIDPAPER, depth: depth, info: screenElementInfo);
    screenData.childViews.add(screenChildElement);
  }

  // This method extract data for Drawer and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "Flow",
  //   "depth": 2,
  //   "info": {
  //    "childrenCount": 7
  //   }
  // }]
  static void extractFlowView(Flow flow, int depth, RetroactiveScreenData screenData) {
    ScreenElementInfo screenElementInfo = ScreenElementInfo(childrenCount: flow.children.length);
    ScreenChildElement screenChildElement = ScreenChildElement(name: FLOW, depth: depth, info: screenElementInfo);
    screenData.childViews.add(screenChildElement);
  }

  // This method extract data for Drawer and place it into
  // retroactiveScreenData as retroactiveScreenId with it's kind
  // example:
  // "childViews": [
  // {
  //   "name": "DataTable",
  //   "depth": 2,
  //   "info": {
  //    "rows": 3,
  //    "columns": 2
  //   }
  // }]
  static void extractDataTableView(DataTable table, int depth, RetroactiveScreenData screenData) {
    ScreenElementInfo screenElementInfo = ScreenElementInfo(columns: table.columns.length, rows: table.rows.length);
    ScreenChildElement screenChildElement = ScreenChildElement(name: DATATABLE, depth: depth, info: screenElementInfo);
    screenData.childViews.add(screenChildElement);
  }

  // This method returns true if the screen name refers to one of the
  // predefined modal kinds
  // "DialogRoute<<of_any_type>>"
  // "_PopupMenuRoute<<of_any_type>>"
  // "CupertinoDialogRoute<<of_any_type>>"
  // "CupertinoModalPopupRoute<<of_any_type>>CupertinoTheme"
  static bool isPageModal(String currentScreenName) {
    if (currentScreenName.contains(DIALOGUE_ROUTE_CANDIDATE) ||
        currentScreenName.contains(POPUP_MENU_ROUTE_CANDIDATE) ||
        currentScreenName.contains(CUPERTINO_DIALOG_CANDIDATE) ||
        currentScreenName.contains(CUPERTINO_MODAL_POPUP_ROUTE_CANDIDATE)) {
      return true;
    }
    return false;
  }
}
