import 'package:flutter/material.dart';
import 'package:flutter/semantics.dart';
// !! we ignore using relative imports only for known issue with logger.dart static filds
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/logger.dart';
import '../constants/constants.dart';
import '../extensions/element_extension.dart';
import '../extensions/map_extension.dart';
import '../extensions/string_extension.dart';
import '../mixins/ui_utils.dart';
import '../utils/common_utils.dart';
import 'tree_abstract.dart';
import 'widget_info.dart';

// This class represent the final element in abstract layout hierarchy that compound the current screen
class ElementInfo extends ITreeElementInfo<WidgetInfo> {
  // Probably not useful as element data because it is runtime dynamically generated
  int _elementHashCode;
  int get elementHashCode => _elementHashCode;

  // Probably not useful as element data because the depth relates to the original
  // tree that very likely to change due to the another widgets in the hierarchy
  int _elementDepth;
  int get elementDepth => _elementDepth;

  // Probably not useful as element data because those type are to general
  String _elementType;
  String get elementType => _elementType;

  Rect _frame;
  Rect get frame => _frame;

  bool _isContainer = false;
  bool get isContainer => _isContainer;

  bool _isClickable = false;
  bool get isClickable => _isClickable;  

  String? _predicate;
  String? get predicate => _predicate;

  String? pendoPredicate;

  String? _accessibilityLabel;
  String? get accessibilityLabel => _accessibilityLabel;

  String? _accessibilityHint;
  String? get accessibilityHint => _accessibilityHint;

  WidgetInfo _widgetInfo;
  WidgetInfo get widgetInfo => _widgetInfo;

  ElementInfo({
    required int elementHashCode,
    required int elementDepth,
    required String elementType,
    required Rect frame,
    required WidgetInfo widgetInfo,
    Element? element,
  })  : _elementHashCode = elementHashCode, _elementDepth = elementDepth, _elementType = elementType, _frame = frame, _widgetInfo = widgetInfo {
    _isContainer = widgetInfo.isContainer;
    _isClickable = widgetInfo.isClickable;
    pendoPredicate = widgetInfo.widgetType;

    //NOTE: pendoPredicate will be a concatenation of the widget types we build in Pendo tree
    // predicate will be a concatenation of X Depth widgets types in Flutter tree
    if (widgetInfo.widgetType == LAYOUT_ID_WIDGET) {
      pendoPredicate = '${pendoPredicate}_${widgetInfo.layoutId}';
    } else if (widgetInfo.widgetType == INDEXED_SEMANTICS_WIDGET) {
      pendoPredicate = '${pendoPredicate}_${widgetInfo.index}';
    }

    // Create predicate for the element - suggest to use the abstract tree hierarchy as predicate
    if (element != null) {
      _predicate = extractPredicateFromElement(element);

      // Semantics
      Map<String, String?> semanticsMap = extractSemanticsFromElement(element);
      _accessibilityLabel = semanticsMap[LABEL]?.base64Str();
      _accessibilityHint = semanticsMap[HINT]?.base64Str();
    }
  }

  ElementInfo.rootElementInfo() : 
    _elementHashCode = 0,
    _elementDepth = 0,
    _elementType = ROOT_VIEW,
    _frame = Rect.fromLTWH(0, 0, Screen.screenWidth, Screen.screenHeight),
    _predicate = ROOT_VIEW,
    pendoPredicate = ROOT_VIEW,
    _widgetInfo = WidgetInfo(PendoRootWidget()),
    _isContainer = true,
    _isClickable = false;

  // This method looking to create only layout basic elements 
  // that contains a screen hierarchy
  static ElementInfo? createLayoutElementInfo(Element element) {
    // If element is not mounted to the current tree, we should skip it
    if (!element.mounted) {
      Logger.e('${element} is not mounted, failed to create tree element');
      return null;
    }

    if (!element.isElementLayoutImportantView()) {
      return null;
    }
    
    return createElementInfo(element);
  } 

  // This method looking to create all kind of elements we 
  // interested to collect into screen hierarchy
  static ElementInfo? createViewElementInfo(Element element) {
    // If element is not mounted to the current tree, we should skip it
    if (!element.mounted) {
      Logger.e('${element} is not mounted, failed to create tree element');
      return null;
    }

    if (!element.isElementView()) {
      return null;
    }
    
    return createElementInfo(element);
  }       

  // This method gather all relevant info from the current Flutter element
  static ElementInfo? createElementInfo(Element element) {
    // If view of interest has no screen position rect, we skip it
    Rect? rect = element.rectInScreenCoordinates();
    if (rect == null) {
      Logger.e('Fail to add element of type: ${element.runtimeType} coz rect is null');
      return null;
    }

    // Create widget info
    WidgetInfo widgetInfo = WidgetInfo(element.widget);
    // Widget can be ignored if it is useless
    if (element.widget is GestureDetector && !widgetInfo.isClickable) {
      return null;
    } else if (element.widget is Ink && widgetInfo.imageName == null) {
      return null;
    }

    // Create general element info
    ElementInfo elementInfo = ElementInfo(
      elementHashCode: element.hashCode,
      elementDepth: element.depth,
      elementType: element.runtimeType.toString(),
      frame: rect,
      widgetInfo: widgetInfo,
      element: element);

    return elementInfo;
  }

  String? retroElementIdentifier() {
    // NOTE: the order should be - hint, identifier, label , text and they should be in base64
    String hashedResult = '';
    
    if (!CommonUtils.isNullOrEmptyString(_accessibilityHint)) {
      hashedResult += 'accessibility.hint:${_accessibilityHint!}';
    }

    // Identifier should go here!
    if (!CommonUtils.isNullOrEmptyString(_accessibilityLabel)) {
      hashedResult += 'accessibility.label:${_accessibilityLabel!}';
    }

    // NOTE: currently we will use only the first text
    String? text = _widgetInfo.getText();
    if (!CommonUtils.isNullOrEmptyString(text)) {
      hashedResult += 'text:${text}';
    }

    if (hashedResult.isEmpty) {
      return null;
    }
    
    return hashedResult.hash256();
  }

  // TODO: Experimental placeholder for predicate extraction, should be discussed further
  String? extractPredicateFromElement(Element element, {int depth = 4}) {
    int localDepth = depth;
    String localPredicate = widgetTypeString(element.widget);
    
    ConditionalElementVisitor visitor = (Element element) {
      if (localDepth == 0) {
        return false;
      }

      localPredicate = '${widgetTypeString(element.widget)}/${localPredicate}';
      localDepth--;
      return true;
    };

    element.visitAncestorElements(visitor);
    return localPredicate;
  }

  @visibleForTesting
  String widgetTypeString(Widget widget) {
    String widgetType = '${widget.runtimeType}';
    if (widget is LayoutId) {
      widgetType = '${widgetType}_${widget.key ?? widget.id}';
    } else if (widget is IndexedSemantics) {
      widgetType = '${widgetType}_${widget.index}';
    }

    return widgetType;
  }

  Map<String, String?> extractSemanticsFromElement(Element element) {
    String? label, hint;
    //TODO: how up through the ancestor tree we should go, maybe we should limit it?
    // TODO: accessibility identifier will be available soon: https://github.com/flutter/flutter/issues/17988
    ConditionalElementVisitor visitor = (Element element) {
      if (element.widget is Semantics) {
        // we assume that the first ancestor semantics is the correct one
        SemanticsProperties properties = (element.widget as Semantics).properties;
        label = properties.label;
        hint = properties.hint;
        return false;
      }

      return true;
    };

    element.visitAncestorElements(visitor);
    return {LABEL: label, HINT: hint};
  }

  @override
  Map<String, dynamic> toJson() {
    Map<String, dynamic> map = {};
    Map<String, String> elementAccessibility = {};
    Map<String, String> textAccessibility = {};
    bool hasAccessibility = false;

    if (!CommonUtils.isNullOrEmptyString(_accessibilityLabel)) {
      hasAccessibility = true;
      elementAccessibility[LABEL] = _accessibilityLabel!.hash256();
      textAccessibility[LABEL_BASE64] = _accessibilityLabel!;
    }

    if (!CommonUtils.isNullOrEmptyString(_accessibilityHint)) {
      hasAccessibility = true;
      elementAccessibility[HINT] = _accessibilityHint!.hash256();
      textAccessibility[HINT_BASE64] = _accessibilityHint!;
    }
    
    map[ZINDEX] = _elementDepth;
    map[ELEMENT_TYPE_NAME] = _elementType;
    map[CLICKABLE] = _isClickable;
    map[IS_CONTAINER] = _isContainer;
    map[POSITION] = {
      LEFT: _frame.left,
      TOP: _frame.top,
      WIDTH: _frame.width,
      HEIGHT: _frame.height,
    };

    String? textBase64 = _widgetInfo.getText();
    map.addIfNotNullOrEmpty(DESC_TEXT_BASE64, textBase64 ?? textBase64Description());
    map.addIfNotNullOrEmpty(TEXT_BASE64, textBase64);

    map[RETRO_ELEMENT_INFO] = <String, dynamic>{
      PREDICATE: _predicate,
      PENDO_PREDICATE: pendoPredicate,
      ACTIONS: _widgetInfo.actions()
    };

    if (hasAccessibility || !CommonUtils.isNullOrEmptyString(textBase64)) {
      (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(ACCESSIBILITY, elementAccessibility); 
      map[RETRO_ELEMENT_TEXTS] = <String, dynamic>{};
      (map[RETRO_ELEMENT_TEXTS] as Map<String, dynamic>).addIfNotNullOrEmpty(ACCESSIBILITY, textAccessibility);
      (map[RETRO_ELEMENT_TEXTS] as Map<String, dynamic>).addIfNotNullOrEmpty(TEXT_BASE64, textBase64);
    }

    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(TEXT, textBase64?.hash256());
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(TYPE, _widgetInfo.widgetType);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(KEY, _widgetInfo.key);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(LAYOUT_ID, _widgetInfo.layoutId?.toString());
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(ICON_CODE, _widgetInfo.iconCode);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(IMAGE_NAME, _widgetInfo.imageName);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(INDEX, _widgetInfo.index);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(COLUMNS, _widgetInfo.columns);
    (map[RETRO_ELEMENT_INFO] as Map<String, dynamic>).addIfNotNullOrEmpty(ROWS, _widgetInfo.rows);

    String? identifier = retroElementIdentifier();
    if (!CommonUtils.isNullOrEmptyString(identifier)) {
      map[RETRO_ELEMENT_HASHES_COLLECTION] = [
        identifier
      ];
    }

    //Currently we will exclude the widgetInfo from elementInfo (retroElementInfo):
    //1. BE/FE don't know how to work with nested maps
    //2. WidgetInfo contains some dynamic elements we should revisit and avoid text by mistake
    //3. WidgetInfo contains some of the elements that elementInfo has
    //4. Every feature that is added to retroElementInfo should have extra processing on FE
    map[WIDGET_INFO] = _widgetInfo.toJson();
    return map;
  }

  String textBase64Description() {
    if (_accessibilityLabel != null) {
      return _accessibilityLabel!;
    }

    if (_accessibilityHint != null) {
      return _accessibilityHint!;
    }
    return this._widgetInfo.widgetType.base64Str() ?? '';
  }

  void squashChild(ElementInfo child) {
    if (CommonUtils.isNullOrEmptyString(_accessibilityHint) && !CommonUtils.isNullOrEmptyString(child._accessibilityHint)) {
      _accessibilityHint = child._accessibilityHint;
    }

    if (CommonUtils.isNullOrEmptyString(_accessibilityLabel) && !CommonUtils.isNullOrEmptyString(child._accessibilityLabel)) {
      _accessibilityLabel = child._accessibilityLabel;
    }

    widgetInfo.squashChild(child.widgetInfo);
  }

  @override
  String toString({String indent = ''}) {
    var lineIndent = '--> ' + indent;
    final buffer = StringBuffer();
    buffer.writeln(lineIndent + toJson().toString());
    return buffer.toString();
  }
}
