// !! we ignore using relative imports only for known issue with logger.dart static filds
// ignore: prefer_relative_imports
import 'package:pendo_sdk/src/logger.dart';
import '../constants/constants.dart';
import 'element_info.dart';
import 'tree_abstract.dart';
import 'tree_node.dart';

typedef Depth = int;
typedef DepthToElementsListMap = Map<Depth, List<TreeNode>>;
typedef HashToElementMap = Map<int, TreeNode>;

class PendoViewHierarchyTree implements ITree<TreeNode> {
  late TreeNode _root;
  TreeNode get root => _root; 

  DepthToElementsListMap _depthToElementsList = {};
  DepthToElementsListMap get depthToElementsList => _depthToElementsList;
  
  List<int> _depthStack = [];
  List<int> get depthStack => _depthStack;
  
  late bool _squashNestedClickableElements;
  bool get squashNestedClickableElements => _squashNestedClickableElements;

  PendoViewHierarchyTree({TreeNode? root, bool squashNestedClickableElements = false}) {
    _root = root ?? TreeNode(elementInfo: ElementInfo.rootElementInfo());
    _squashNestedClickableElements = squashNestedClickableElements;
    updateTrackedElementsDepth(_root);
  }

  void updateDepth(int currentElementDepth) {
    while ( _depthStack.length > 0 && currentElementDepth <= _depthStack.last) {
      _depthStack.removeLast();
    }
  }

  @override
  addTreeElement(TreeNode treeElement) {
    // this always returns a direct parent for the current treeElement. 
    // if the result is null, treeElement is then a sibling of the direct parent.
    TreeNode? parentTreeElement = _depthToElementsList[_depthStack.last]?.last;
    
    if (parentTreeElement == null) {
      Logger.e('Parent element should not be null');
      return;
    }

    // if our parent and current treeElement is clickable and current treeElement is contained in the other, we will skip
    // the current treeElement to avoid multiple clickable elements or any other duplicate elements in the future
    if (_squashNestedClickableElements && treeElement.isNestedClickableElement(parentTreeElement)) {
      return;
    }

    // if parent is clickable and current treeElement is not clickable - we will squash its texts to its parent
    // this comes to enable texts on clickable elements and to allow better feature rules on FE, like exist on native side.
    if (treeElement.shouldSquashChild(parentTreeElement)) {
      parentTreeElement.elementInfo.squashChild(treeElement.elementInfo);
      Logger.d('squashed $treeElement');
      return;
    }

    parentTreeElement.addChild(treeElement);
    updateChildPredicate(parentTreeElement, treeElement);
    updateTrackedElementsDepth(treeElement);
  }

  void updateTrackedElementsDepth(TreeNode treeElement) {
    if (!_depthToElementsList.containsKey(treeElement.elementDepth)) {
      _depthToElementsList[treeElement.elementDepth] = [];
    }
    _depthToElementsList[treeElement.elementDepth]?.add(treeElement);
    _depthStack.add(treeElement.elementDepth);
  }

  void updateChildPredicate(TreeNode parentTreeElement, TreeNode treeElement) {
    // adjust child predicate based on it's parent
    treeElement.elementInfo.pendoPredicate = '${parentTreeElement.elementInfo.pendoPredicate}/${treeElement.elementInfo.pendoPredicate}';
  }

  @override
  String toString() => _root.toString();

  @override
  Map<String, dynamic> toJson() {
    Map<String, dynamic> map = {};
    map[ROOT] = _root.toJson();
    return map;
  }
}
